from inflection import dasherize, underscore

from drf_shortcuts.views import create_standard_viewset_class


def register_standard_endpoint(router, model, viewset_cls=None):
    endpoint_name = dasherize(underscore(model.__name__))
    viewset = viewset_cls or create_standard_viewset_class(model)
    router.register(endpoint_name, viewset, base_name=viewset.serializer_class.DEFAULT_BASE_NAME)



from rest_framework import serializers
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework.fields import empty
from django.conf import settings
import re
from inflection import dasherize, underscore


def generate_serializer_base_name(model_cls):
    return dasherize(underscore(model_cls.__name__))


def generate_detail_view_name(base_name):
    default_detail_name = base_name + "-detail"
    if settings.API_URL_NAMESPACE:
        return "{}:{}".format(settings.API_URL_NAMESPACE, default_detail_name)
    return default_detail_name


def rename_serializer_field(serializer, source_name, target_name, display_name=None):
    field = serializer.fields.pop(source_name)
    if display_name is not None:
        field.label = display_name
    serializer.fields[target_name] = field


def get_entity_pk(serializer):
    view_kwargs = serializer.context['view'].kwargs if 'view' in serializer.context else {}
    return view_kwargs.get('pk') if 'pk' in view_kwargs else None


def get_required_field_value(data, field_name, serializer, fetch_model):
    if field_name in data:
        return data.get(field_name)
    pk = get_entity_pk(serializer)
    assert pk is not None, "Update or partial update is assumed"
    field_value = getattr(fetch_model(pk), field_name)
    assert field_value is not None, "Unexpectedly required field value is None"
    return field_value


def get_optional_field_value(data, field_name, serializer, fetch_model):
    if field_name in data:
        return data.get(field_name)
    pk = get_entity_pk(serializer)
    return None if pk is None else getattr(fetch_model(pk), field_name)


class OptimizeUrlFieldsSerializer(serializers.ModelSerializer):
    explicit_url_field_names = []

    def __init__(self, instance=None, data=empty, **kwargs):
        super().__init__(instance, data, **kwargs)
        remove_urls = True
        if 'request' in self.context:
            query_params = self.context['request'].query_params
            if 'forceUrls' in query_params:
                remove_urls = query_params.get('forceUrls') == 'false'
            else:
                remove_urls = not isinstance(self.context['request'].accepted_renderer, BrowsableAPIRenderer)
        if remove_urls:
            field_names_to_remove = []
            for field_name in self.fields.keys():
                is_id_field = isinstance(self.fields[field_name], serializers.HyperlinkedIdentityField)
                is_related_field = isinstance(self.fields[field_name], serializers.HyperlinkedRelatedField)
                is_explicit_field = field_name in self.explicit_url_field_names
                if is_id_field or is_related_field or is_explicit_field:
                    field_names_to_remove.append(field_name)
            for field_name in field_names_to_remove:
                del self.fields[field_name]


class ValidateAndInsertAuthorSerializer(serializers.ModelSerializer):
    author_field_name = None

    _author_inserted = False

    def validate(self, data):
        author = self._get_author(data)
        if author is not None and not self._author_inserted:
            user = self.context['request'].user
            if user != author:
                raise serializers.ValidationError('Authoring for another user is not allowed')
        return data

    def to_internal_value(self, data):
        result = super().to_internal_value(data)
        if self.context is not None and 'request' in self.context and self.context['request'].method == 'POST':
            author = self._get_author(result)
            if author is None and self.context.get('request') is not None:
                user = self.context['request'].user
                if self.author_field_name:
                    result[self.author_field_name] = user
                else:
                    self.set_author_core(result, user)
                self._author_inserted = True
        return result

    def _get_author(self, data):
        if self.author_field_name:
            return data.get(self.author_field_name)
        else:
            return self.get_author_core(data)

    def get_author_core(self, data):
        raise NotImplementedError("This should be implemented")

    def set_author_core(self, data, author):
        raise NotImplementedError("This should be implemented")


class UpdateEditorSerializer(serializers.ModelSerializer):
    def to_internal_value(self, data):
        result = super().to_internal_value(data)
        in_request_context = self.context is not None and 'request' in self.context
        request = self.context['request'] if in_request_context else None
        if in_request_context and request.method in ('PUT', 'PATCH') and request.user:
            self.set_editor_core(result, request.user)
        return result

    def set_editor_core(self, data, editor):
        raise NotImplementedError("This should be implemented")


class JsFriendlyFieldsRenamingSerializer(serializers.ModelSerializer):
    regex = re.compile('_(.)')

    def __init__(self, instance=None, data=empty, **kwargs):
        super().__init__(instance, data, **kwargs)
        for field_name in [f for f in self.fields if '_' in f]:
            rename_serializer_field(self, field_name, self.regex.sub(self._process_match, field_name))

    @staticmethod
    def _process_match(match):
        return match.group(1).upper()


def create_standard_serializer_class(model_cls):
    base_name = generate_serializer_base_name(model_cls)

    class Serializer(OptimizeUrlFieldsSerializer, JsFriendlyFieldsRenamingSerializer):
        DEFAULT_BASE_NAME = base_name

        url = serializers.HyperlinkedIdentityField(view_name=generate_detail_view_name(DEFAULT_BASE_NAME))

        class Meta:
            model = model_cls
            fields = '__all__'

    Serializer.__name__ = model_cls.__name__ + 'Serializer'
    return Serializer



from rest_framework.settings import api_settings
from rest_framework import viewsets
from django.db.models import TextField, CharField
from django.db.models.fields.related import ForeignObjectRel
from inflection import pluralize, humanize, underscore

from drf_shortcuts.serializers import create_standard_serializer_class


def append_pagination_info_to_docstring(cls):
    if cls.__doc__ is not None:
        cls.__doc__ = '{}\n' \
                      'Specify "?page=<page number>" to get particular page. ' \
                      'Page size is {}.\n'.format(cls.__doc__, api_settings.PAGE_SIZE)
    return cls


def append_search_info_to_docstring(cls):
    if cls.__doc__ is not None:
        cls.__doc__ = '{}\n' \
                      'Specify "?search=<search terms here>" query parameter to search items.\n' \
                      .format(cls.__doc__)
    return cls


def append_ordering_info_to_docstring(fields):
    assert len(fields) > 0, "At least one ordering field is required"

    def _wrapped_append(cls):
        if cls.__doc__ is not None:
            appended_doc = 'Specify "?ordering=<fields to order by here>" query parameter to order results.\n\n' \
                           'You can use following fields for ordering: {}.\n\n' \
                           "To reverse ordering of a field prefix it with hyphen '-': ?ordering=-{}.\n" \
                           .format(', '.join(fields), fields[0])
            if len(fields) > 1:
                appended_doc = '{}' \
                               'You can specify multiple orderings by separating them using comma: ?ordering={}.\n' \
                               .format(appended_doc, ','.join(fields[:2]))
            cls.__doc__ = '{}\n{}'.format(cls.__doc__, appended_doc)
        return cls

    return _wrapped_append


def append_search_ordering_and_pagination_info_to_docstring(fields):
    def _wrapped_append(cls):
        return append_pagination_info_to_docstring(
            append_ordering_info_to_docstring(fields)(
                append_search_info_to_docstring(cls)))

    return _wrapped_append


def get_fields_suitable_for_ordering(model):
    return [f.name for f in model._meta.get_fields()
            if not (isinstance(f, TextField) or isinstance(f, ForeignObjectRel))]


def get_fields_suitable_for_search(model):
    return [f.name for f in model._meta.get_fields() if isinstance(f, CharField)]


def create_standard_viewset_class(model_cls, serializer_cls=None):
    fields = get_fields_suitable_for_ordering(model_cls)

    class ViewSet(viewsets.ModelViewSet):
        serializer_class = serializer_cls or create_standard_serializer_class(model_cls)
        search_fields = get_fields_suitable_for_search(model_cls)
        ordering_fields = fields
        lookup_value_regex = '[^/]+'

        def get_queryset(self):
            return model_cls.objects.all().order_by('id')

    plural_model = pluralize(model_cls.__name__)
    ViewSet.__name__ = plural_model + 'ViewSet'
    ViewSet.__doc__ = 'This is {} API endpoint.\n'.format(humanize(underscore(plural_model)).lower())
    append_search_ordering_and_pagination_info_to_docstring(fields)(ViewSet)
    return ViewSet



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

import drf_shortcuts

setup(
    name='drf_shortcuts',
    version=drf_shortcuts.__version__,
    description="Common shortcuts for speeding up your development based on Django REST Framework (DRF).",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Daniel Ivanov',
    author_email='megaden4eg@gmail.com',
    url='https://github.com/megaden/drf-shortcuts',
    license='MIT',
    packages=['drf_shortcuts'],
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=['inflection>=0.3'],
)
